import os
import numpy as np
import tensorflow as tf


def conv_tf(weight, x, strides, padding):
    conv_out = tf.nn.depthwise_conv2d(
        input=x,
        filter=weight,
        strides=strides,
        padding=padding,
    )
    return conv_out

def gen_data(channel_in, kernel_sizes, input_h, input_w, dtype):
    """Generate input and weight"""
    weight = tf.cast(
        tf.random.uniform((kernel_sizes[0], kernel_sizes[1], channel_in, channel_multiplier)),
        dtype=dtype)
    x = tf.cast(tf.random.uniform((1, input_h, input_w, channel_in)), dtype=dtype)

    # save data
    wf = open("weight.txt", 'w')
    weight_size = kernel_sizes[0] * kernel_sizes[1] * channel_in * channel_multiplier
    print('weight size', weight_size)
    # transpose [k_h, k_w, in_c, out_c] to [out_c, in_c, k_h, k_w] for CUDNN
    weight_np = weight.numpy().transpose(3, 2, 0, 1)
    weight_np = np.reshape(weight_np, (weight_size, ))
    for i in range(weight_size):
        wf.write('%f ' % weight_np[i])
    wf.close()

    wf1 = open("input.txt", 'w')
    x_size = input_h * input_w * channel_in
    # transpose input to [batch, in_channel, height, width]
    x_np = x.numpy().transpose(0, 3, 1, 2)
    print('x_size', x_size)
    x_np = np.reshape(x_np, (x_size, ))
    for i in range(x_size):
        wf1.write('%f ' % x_np[i])
    wf1.close()

    return weight, x

if __name__ == "__main__":
    # define hyper parameters
    channel_out = 8
    channel_in = 8
    kernel_sizes = (3, 3)
    strides = (1, 1)
    channel_multiplier = channel_out // channel_in
    dtype = tf.float16
    input_sizes = (4, 4)
    weight, x = gen_data(channel_in, kernel_sizes, input_sizes[0], input_sizes[1], dtype)

    # initialize weight and input
    tf_strides = [1, strides[0], strides[1], 1]
    padding = "SAME"

    # get tensorflow output
    conv_out_tf = conv_tf(weight, x, tf_strides, padding)
    print("---------- TF output ---------")
    print(conv_out_tf)

    # compile C++ program w/ cudnn
    os.system('nvcc conv_depthwise.cpp -o conv_depthwise -I /usr/local/cuda-11.2/include -L /usr/local/cuda-11.2/lib64 -lcudnn -lcudart')

    # get CUDNN output
    command = './conv_depthwise %s %s %s %s %s %s %s' % \
        (channel_in, input_sizes[0], input_sizes[1], channel_out, kernel_sizes[0], strides[0], channel_out)
    os.system(command)

